import osmnx as ox
import networkx as nx
#import folium
import itertools
import matplotlib.pyplot as plt


G=nx.MultiGraph()
edges=[(0,1,1),(1,2,1),(2,3,1),(3,4,1),(4,5,1),(5,6,1),(6,7,1),
       (8,3,2),(9,4,1),(9,5,1),(10,7,2),
       (8,9,2),(9,10,2),
       (8,11,2),(12,10,2),
       (11,12,2),(12,13,2),(13,7,2),
       (11,14,2),(14,15,4),(15,13,1)]
G.add_weighted_edges_from(edges,'length')
pos=[(0,4),(1,4),(2,4),(3,4),(4,4),(5,4),(6,4),(7,4),
     (1,3),(3,3),(5,3),
     (3,2),(6,2),(8,2),
     (4,1),(7,1)]

# uncomment for adjacent 4nodes testing
#G.add_weighted_edges_from( [(10,6,0.1)],'length') # test adjacent 4-nodes

# uncomment for parallel edges testing
#G.add_weighted_edges_from( [(14,15,1)],'length') # test parallel edges

# uncomment for parallel edges testing involving 4node pairs
#G.add_weighted_edges_from( [(7,13,1)],'length') # test parallel edges

# uncomment for parallel edges between 4node and a 3node
G.add_weighted_edges_from( [(13,15,1)] ,'length') #test parallel edges

# select between new and old algorithms
new_algo = True
new_algo = False



def draw(Q, title=''):
    ax = plt.gca()
#    nx.draw_networkx_nodes(Q, pos, **options)
#    nx.draw_networkx(Q, pos, **options)
    options = {
        "font_size": 16,
        "node_size": 1000,
        "node_color": "white",
        "edgecolors": "black",
        "linewidths": 5,
        "width": 5,
    }
#    nx.draw(Q, pos, with_labels=True,
#            onnectionstyle='arc3, rad = 0.1',
#            **options)
  
    
    for e in Q.edges:
       ax.annotate("",
               xy=pos[e[0]], xycoords='data',
               xytext=pos[e[1]], textcoords='data', zorder=0,
               arrowprops=dict(arrowstyle="-", color="0.5",
                               shrinkA=5, shrinkB=5,
                               patchA=None, patchB=None,
                               connectionstyle="arc3,rad=rrr".replace('rrr',str(0.3*e[2])
                               ),
                               ),
               )
    nx.draw_networkx_nodes(Q, pos, node_color = "white", node_size = 1000,
                           edgecolors='black',alpha=1,linewidths = 3)
    nx.draw_networkx_labels(Q, pos, font_size=12, font_color='k')
    

    plt.axis('off')
    plt.title(title,fontsize='large',loc='left',fontweight='bold',style='italic',family='monospace')
    plt.show()    


  
def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('')
        
report(G)
draw(G)

        


# remove trailheads and termini
def delete_ends(Q):
    while True: # simulate do-while
        rn = []
        for node, neighbors in Q.adjacency():
            if len(neighbors) == 1:
                rn.append(node)
        if len(rn)>0:            
            Q.remove_nodes_from(rn)
        else:
            break

delete_ends(G)
report(G,'in undirected multigraph with ends removed')
#draw(G,'undirected multigraph with ends removed')

# remove nodes that only have two neighbor nodes
def contract_twos(Q):
    while True: # simulate do-while
        rn = []
        for node, neighbors in Q.adjacency():
            #if len(neighbors) == 2:
            if Q.degree(node) ==2:
               neighbor = next(iter(neighbors))
               rn.append(node)
               rn.append(neighbor)
               break
        if len(rn)==0:
            break;
        else:  # do not change the graph within a for loop
            nx.contracted_nodes(Q, node, neighbor, self_loops=False, copy=False)
            
contract_twos(G)
report(G,'with two-nodes merged')
draw(G,'two-nodes merged')

def odd_nodes_graph(Q):
    R = nx.MultiGraph()
    for u, v, d in Q.edges(data=True):
        if Q.degree(u) % 2 == 1 and Q.degree(v) % 2 == 1:
            R.add_edge(u,v,**d)
    return R




def fix_covalent_4nodes(Q):
    QR = Q.copy()
    for u, v, d in Q.edges(data=True):
        if Q.degree(u) % 2 == 0 or Q.degree(v) % 2 == 0:
            edge_count = Q.number_of_edges(u,v)
            if edge_count==2:
                if QR.has_edge(u,v,1):
                    QR.remove_edge(u,v,1) #fix: find smallest edge
    return QR
    
if new_algo or True:
    G = fix_covalent_4nodes(G)
    contract_twos(G)
    draw(G,'covalents fixed')
K = odd_nodes_graph(G)           
report(K,'in odd-nodes graph')
draw(K,'odd-nodes graph')

# add in edges that bypass 4-nodes in graph
# fix: 4-way connected to a 4-way
def complete_3node_graph(Q,R):
    n4list = []
    for node, neighbors in Q.adjacency():
        if len(neighbors) == 4:
            n4list.append( (node,neighbors) )
            nlist = []
            llen = []
            for neighbor, edict in neighbors.items():
                w = min([d['length'] for d in edict.values()])
                nlist.append(neighbor)
                llen.append(w)
              #  for dic in edict.items():
                    #w = dic['length']
                    #nlist.append(neighbor)
                    #llen.append(w)
            # could use itertools?
            for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:
                R.add_edge(nlist[i],nlist[j], length= llen[i]+llen[j]) 
    return n4list

def complete_3node_graph_new(Q,R): #original, 3node
    QR = R.copy()
    n4set = set()
    n4pairs = set()
    n4neighbors = set()

# get set of 4nodes
# For this next section, we could merge for loops for efficiency,
# but separate is clearer
    for node in Q:
        d = Q.degree(node)
        if d>2 and d%2==0:  # well, we had a 6 way intersection
            n4set.add(node)
    # get set of 4nodes with 4node neighbors
    for node in n4set:
        for neighbor in Q.neighbors(node):
            if (neighbor in n4set) and (neighbor != node):
                n4pairs.add( ( min(node,neighbor),max(node,neighbor) ) )

    print('n4set: ', n4set)
    print('n4pairs: ', n4pairs)
                
    # add bridging edges for n4set
    for node in n4set:
        list_edges = []
        for u,v,k,d in Q.edges(node,keys=True,data='length',default=0):
            # remove the edge case of 4node connected to a 4node
            if (u not in n4set) :    
    #        if (u not in n4set) and (u not in n4neighbors):    
                list_edges.append( (v,u,k,d) )
                n4neighbors.add(u)
            elif (v not in n4set) :
    #        elif (v not in n4set) and (v not in n4neighbors):
                list_edges.append( (u,v,k,d) )
                n4neighbors.add(v)
        for e1,e2 in itertools.combinations(list_edges,2):
            QR.add_edge(e1[1],e2[1],length = e1[3]+e2[3], node4_path= [ e1[1],node,e2[1] ] ) #fix: add geometry for draw()
    print('number of 4nodes bridged: ', len(n4set))

    # in add bridging edges for n4pairs, somewhat more complicated
    for node,node2 in n4pairs:
        list_edges = []
        list_edges2 = []
        for u,v,k,d in Q.edges(node,keys=True,data='length',default=0):
        # remove the edge case of 4node connected to a 4node
            if u not in n4set:    # first node in tuple is 4node
                list_edges.append( (v,u,k,d) )
            elif v not in n4set:
                list_edges.append( (u,v,k,d) )
        for u,v,k,d in Q.edges(node2,keys=True,data='length',default=0):
        # remove the edge case of 4node connected to a 4node
            if u not in n4set:    
                list_edges2.append( (v,u,k,d) )
            elif v not in n4set:
                list_edges2.append( (u,v,k,d) )
        d = Q.edges[node,node2,0]['length'] #fix: parallel edges for bridge
        if Q.number_of_edges(node,node2)!=1:
            print('parallel edges between 4nodes ', node, ' ', node2)
        bridge = (node,node2,0,d) #fix edge case multiple edges between node,node2
        for e1 in list_edges:
            for e2 in list_edges2:
                QR.add_edge(e1[1],e2[1],length = e1[3]+e2[3]+bridge[3], node4_path=[ e1[1],node,node2,e2[1] ]) #fix: add geometry for draw()
    print('number of 4node-pairs bridged: ', len(n4pairs))
    return QR

k_fix_4_list = []
if new_algo:
    K = complete_3node_graph_new(G,K)
else:
    k_fix_4_list = complete_3node_graph(G,K)

report(K,'in odd-nodes 4node bypass')
draw(K,'odd-nodes 4node bypass')

#
 
# get a minimum matching of odd nodes
odd_matches_set = nx.algorithms.min_weight_matching(nx.Graph(K),'length')
# perhaps a matching is not possible?
# Might have to do my own algorithm?
print('number of odd matches', len(odd_matches_set))

# delete matched edges
# do I need to de-duplicate odd_matches_list?

def remove_first_edges_from(Q,edges):
    QR = Q.copy()
    for u,v in edges:
        for k in range(9):
            if QR.has_edge(u,v,k):
                QR.remove_edge(u,v,k)
                break
    return QR

if new_algo:
    K = remove_first_edges_from(K,odd_matches_set)
else:
    K.remove_edges_from(odd_matches_set)

report(K,'with odd-node matches removed')
#draw(K,'odd-node matches removed')



# how do I fix the 4-nodes?
def restore_4node_graph(Q,fix_4_list):
    for node, neighbors in fix_4_list:
       i = 0
       nlist = []
       vlist = []
       for neighbor, edict in neighbors.items():
           # for dic in edict:
           nlist.append(neighbor)
           vlist.append(edict[0])
       for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:               
            if Q.has_edge(nlist[i],nlist[j]):
                Q.remove_edge(nlist[i],nlist[j])
                if not Q.has_edge(node,nlist[i]):
                    Q.add_edge(node,nlist[i],**vlist[i]) #fix
                if not Q.has_edge(node,nlist[j]):                                  
                    Q.add_edge(node,nlist[j],**vlist[j])

def restore_4node_graph_new(Q,Q_original):
    QR = Q.copy()    
    for u,v,k,dic in Q.edges(data=True, keys=True):
        if 'node4_path' in dic:
            QR.remove_edge(u,v,key=k)
            path = dic['node4_path']
            # subgraph and compose do not seem to be working for me, so do explicitly
            for w,x in zip(path[:-1],path[1:]): #takes two nodes at a time
                if not QR.has_edge(w,x):
                    for j in Q_original[w][x]:
                        dat = Q_original.get_edge_data(w,x,j)
                        QR.add_edge(w,x,j,**dat)
                        break #fix restore longest edge, not first edge?
##    n4set = set()
##    n4pairs = set()
##    #go ahead and regenerate n4set and n4pairs, stay modular
##    for node in Q_original:
##        d = Q_original.degree(node)
##        if d>2 and d%2==0:  # well, we had a 6 way intersection
##            n4set.add(node)
##    # get set of 4nodes with 4node neighbors
##    for node in n4set:
##        for neighbor in Q_original.neighbors(node):
##            if (neighbor in n4set) and (neighbor != node):
##                n4pairs.add( ( min(node,neighbor),max(node,neighbor) ) )
##
##    print('n4set: ', n4set)
##    print('n4pairs: ', n4pairs)
##                     

##    n4pair_restore_count = 0
##    for u,v in n4pairs:
##        if QR.degree(u) %2==1 and QR.degree(v)%2==1:
##        #fix: handle case of deleted node
##            if QR.has_edge(u,v):  #fix: try else first
##                QR.remove_edge(u,v)
##                n4pair_restore_count +=1
##            else:
##                for k, dat in Q_original.edges[u][v]:
##                    QR.add_edge(u,v,k,dat)
##                    n4pair_restore_count +=1
##                    break
##    if n4pair_restore_count>0:
##        print('number of n4pairs adjusted: ', n4pair_restore_count)
    #restore node attributes if needed
    restore_count = 0
    for node in path:  # add node attributes back in
        node_dic = QR.nodes[node]
        if 'x' not in node_dic:
            node_dic = Q_original.nodes[node]
            nx.set_node_attributes(QR,node_dic)
            restore_count += 1
    if restore_count > 0:
        print('number of 4nodes restored: ', restore_count)
    #check_node_geometry(QR)       
    return QR                                

if new_algo:
    K = restore_4node_graph_new(K,G)
else:
    restore_4node_graph(K,k_fix_4_list)

report(K,'odd-matches deleted with 4nodes restored')                             
draw(K,'odd-matches deleted with 4nodes restored')


# we expect some disconnected subgraphs, so choose the largest
nodes = max(nx.connected_components(K))
L = nx.subgraph(K, nodes)
report(L,'in largest subgraph with odd-node matches removed')

if nx.is_eulerian(L):
    print('L is eulerian')
    print('The node path is: ', [u for u,v in nx.eulerian_circuit(L)])
            

#FIX: Do random graph connections on a small graph, check errors, iterate.




